/*
 ATmega VFD
 Version: 0.1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 1/08/2014
 Last Modified: 1/08/2014
 
 Use for testing Noritake 320x32 VFD - MN32032A
 
 Remember to change your fuse bits so the ATmega328P uses a 20MHz crystal
 avrdude -p atmega328p -c usbtiny -U lfuse:w:0xde:m -U hfuse:w:0xd9:m -U efuse:w:0xff:m
 
 */

// ATmega88 Pin Map
//
//               +-\/-+
//   Reset PC6  1|o   |28  PC5 SIN2
//         PD0  2|    |27  PC4 SIN1
//         PD1  3|    |26  PC3 CLK
//         PD2  4|    |25  PC2 GCP
//         PD3  5|    |24  PC1 LAT
//         PD4  6|    |23  PC0 BLK
//         VCC  7|    |22  GND
//         GND  8|    |21  AREF
//    XTAL PB6  9|    |20  AVCC
//    XTAL PB7 10|    |19  PB5 SCK
//         PD5 11|    |18  PB4 MISO
//         PD6 12|    |17  PB3 MOSI
//         PD7 13|    |16  PB2
//         PB0 14|    |15  PB1
//               +----+

#define F_CPU 20000000 // 20 MHz

#define BLK PC0
#define LAT PC1
#define GCP PC2
#define CLK PC3
#define SIN1 PC4
#define SIN2 PC5

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "setup.c"

const uint8_t ledLetters[26][32] PROGMEM = {
	{0,0,0,0,31,96,0,63,64,128,128,131,124,
	0,0,0,0,0,128,128,128,128,128,128,128,224}, // a
	
	{224,32,32,32,39,40,48,32,32,32,48,40,231,
	0,0,0,0,128,96,32,16,16,16,32,64,128}, // b
	
	{0,0,0,0,62,65,128,128,128,128,128,65,62,
	0,0,0,0,128,128,128,0,0,0,128,0}, // c
	
	{1,0,0,0,30,97,128,128,128,128,64,97,30,
	192,64,64,64,64,64,192,64,64,64,192,192,112}, // d
};

uint8_t grid = 0;
uint8_t gridblank = 0;
uint8_t LEDBuf[40][32]; 
uint8_t LEDBufHasData[80];

void shiftout(uint8_t data) {
	PORTC &= ~(1<<CLK);	
	
	for (int8_t x = 7; x >= 0; x--) {
		if (data & (1<<x)) {
			PORTC |= (1<<SIN1);
		}
		else {
			PORTC &= ~(1<<SIN1);
		}
		
		PORTC |= (1<<CLK);
		if (x > 0) {
			PORTC &= ~(1<<CLK);
		}
	}
}

int main(void) {
	setup();
	
	for (uint8_t ledx = 0; ledx < 40; ledx++) {
		for (uint8_t ledy = 0; ledy < 32; ledy++) {
			LEDBuf[ledx][ledy] = 0;
		}
	}
	
	for (uint8_t ledy = 0; ledy < 13; ledy++) {
		LEDBuf[0][ledy] = (unsigned char) pgm_read_byte(&ledLetters[0][ledy]);
		LEDBuf[1][ledy] = (unsigned char) pgm_read_byte(&ledLetters[0][ledy+13]);
		LEDBuf[2][ledy] = (unsigned char) pgm_read_byte(&ledLetters[1][ledy]);
		LEDBuf[3][ledy] = (unsigned char) pgm_read_byte(&ledLetters[1][ledy+13]);
		LEDBuf[4][ledy] = (unsigned char) pgm_read_byte(&ledLetters[2][ledy]);
		LEDBuf[5][ledy] = (unsigned char) pgm_read_byte(&ledLetters[2][ledy+13]);
		LEDBuf[6][ledy] = (unsigned char) pgm_read_byte(&ledLetters[3][ledy]);
		LEDBuf[7][ledy] = (unsigned char) pgm_read_byte(&ledLetters[3][ledy+13]);
		LEDBuf[8][ledy] = (unsigned char) pgm_read_byte(&ledLetters[0][ledy]);
		LEDBuf[9][ledy] = (unsigned char) pgm_read_byte(&ledLetters[0][ledy+13]);
		LEDBuf[10][ledy] = (unsigned char) pgm_read_byte(&ledLetters[1][ledy]);
		LEDBuf[11][ledy] = (unsigned char) pgm_read_byte(&ledLetters[1][ledy+13]);
		LEDBuf[12][ledy] = (unsigned char) pgm_read_byte(&ledLetters[2][ledy]);
		LEDBuf[13][ledy] = (unsigned char) pgm_read_byte(&ledLetters[2][ledy+13]);
		LEDBuf[14][ledy] = (unsigned char) pgm_read_byte(&ledLetters[3][ledy]);
		LEDBuf[15][ledy] = (unsigned char) pgm_read_byte(&ledLetters[3][ledy+13]);
	}
	
	for (int e = 0; e < 30; e++) {
		LEDBufHasData[e] = 1; 
	}

	uint8_t startup = 0;
	while (1) {
		if (startup < 200) {
			PORTC |= (1<<BLK);
			startup++;
			_delay_ms(2);
		}
		else {
			PORTC &= ~(1<<BLK);
		}
		
		uint8_t counter = 0;
		
		// Check which grids to light
		while (LEDBufHasData[grid] != 1) {
			grid++;
			if (grid == 80) {
				grid = 0;
				break;
			}
		}
		
		for (counter = 0; counter < 13; counter++) {
			uint8_t datawrite = 0;
			
			if (grid % 2 == 0) {
				if (LEDBuf[(grid/2)][counter] & 1<<7) {
					datawrite += 1<<7; // a
				}
				if (LEDBuf[(grid/2)][counter] & 1<<6) {
					datawrite += 1<<5; // b
				}
				if (LEDBuf[(grid/2)][counter] & 1<<5) {
					datawrite += 1<<3; // c
				}
				if (LEDBuf[(grid/2)][counter] & 1<<4) {
					datawrite += 1<<1; // d
				}
			}
			else {
				if (LEDBuf[(grid/2)][counter] & 1<<3) {
					datawrite += 1<<0; // e
				}
				if (LEDBuf[(grid/2)][counter] & 1<<2) {
					datawrite += 1<<2; // f
				}
				if (LEDBuf[(grid/2)][counter] & 1<<1) {
					datawrite += 1<<4; // g
				}
				if (LEDBuf[(grid/2)][counter] & 1<<0) {
					datawrite += 1<<6; // h
				}
			}
			shiftout(datawrite);
		}
		
		for (counter = 0; counter < 15; counter++) {
			shiftout(0);
		}
		
		PORTC |= (1<<GCP);
		_delay_loop_1(2);
		PORTC &= ~(1<<GCP);
		_delay_loop_1(2);
		
		for (counter = 0; counter < 4; counter++) {
			shiftout(0);
		}
		
		PORTC |= (1<<GCP);
		_delay_loop_1(2);
		PORTC &= ~(1<<GCP);
		_delay_loop_1(2);
		
		// Grids		
		int8_t gridcheck = 80;
		for (int8_t x = 0; x < 10; x++) {
			gridcheck = gridcheck - 8;
			
			if (grid == gridcheck) { // a
				shiftout(1);
			}
			else if (grid == gridcheck+1) { // a
				shiftout(3);
			}
			else if (grid == gridcheck+2) { // a
				shiftout(6);
			}
			else if (grid == gridcheck+3) { // a
				shiftout(12);
			}
			else if (grid == gridcheck+4) { // b 
				shiftout(24);
			}
			else if (grid == gridcheck+5) { // b
				shiftout(48);
			}
			else if (grid == gridcheck+6) { // b 
				shiftout(96);
			}
			else if (grid == gridcheck+7) { // b
				shiftout(192);
			}
			else {
				shiftout(0);
			}
		}
		
		PORTC |= (1<<LAT);
		_delay_loop_1(2);
		PORTC &= ~(1<<LAT);
		
		grid++;
		if (grid == 80) {
			grid = 0;
		}
	}
}
